from random import shuffle
from typing import List, Dict, Tuple, Optional

import numpy as np

__all__ = [
    "StandardGetter"
]


class StandardGetter:

    def __init__(
        self,
        x_names: List[str],
        y_names: List[str],
        x: np.ndarray,
        y: np.ndarray
    ):
        self.x_names = x_names
        self.y_names = y_names
        self.x = x
        self.y = y

    def get(
        self,
        x_features: List[str],
        y_features: List[str],
        restrictions: Dict[str, Tuple[float]],
        max_samples: Optional[int]=None
    ):
        assert set(x_features) <= set(self.x_names)
        assert set(y_features) <= set(self.y_names)

        x_idx = [self.x_names.index(v) for v in x_features]
        _x = np.column_stack([
            self.x[:, i] for i in x_idx
        ])

        y_idx = [self.y_names.index(v) for v in y_features]
        _y = np.column_stack([
            self.y[:, i] for i in y_idx
        ])
        
        filt = np.ones(_y.shape[0], dtype="bool")

        # Remove non-finite pixels in variables
        filt &= np.isfinite(_x).all(axis=1)

        # Remove pixels out of the targets ranges (including NaNs)
        for key, (low, upp) in restrictions.items():
            if key in self.x_names:
                i = self.x_names.index(key)
                filt &= np.logical_and(
                    self.x[:, i] >= (low or -float('inf')),
                    self.x[:, i] <= (upp or float('inf'))
                )
            elif key in self.y_names:
                i = self.y_names.index(key)
                filt &= np.logical_and(
                    self.y[:, i] >= (low or -float('inf')),
                    self.y[:, i] <= (upp or float('inf'))
                )

        _x = _x[filt]
        _y = _y[filt]
        
        if max_samples is None or _x.shape[0] <= max_samples:
            return _x, _y
        
        idx = list(range(_x.shape[0]))
        shuffle(idx)
        idx = idx[:max_samples]

        return _x[idx, :], _y[idx, :]



from typing import Tuple, Union

import numpy as np
from scipy.linalg import sqrtm

__all__ = [
    "contraction_matrix",
    "canonical_corr",
    "cca"
]

def contraction_matrix(X: np.ndarray, Y: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Returns the contraction matrix as well as the matricial square-root of the covariance matrices.
    """
    
    # Number of pixels
    L = X.shape[0]
    
    # Centering X and Y
    Xc = X - np.mean(X, 0)
    Yc = Y - np.mean(Y, 0)

    # Unbiased covariance and crosses covariance matrices of X and Y
    R_XX = Xc.T @ Xc / (L-1)
    R_YY = Yc.T @ Yc / (L-1)
    R_XY = Xc.T @ Yc / (L-1)

    # Matricial square-root of covariance matrices
    R_XX_Pud = sqrtm(R_XX)
    R_YY_Pud = sqrtm(R_YY)

    # Contraction matrix (which singular values are the canonical correlations)   
    M = np.linalg.solve(R_XX_Pud, R_XY) @ np.linalg.inv(R_YY_Pud)
    
    return M, R_XX_Pud, R_YY_Pud

def canonical_corr(X: np.ndarray, Y: np.ndarray, max: bool=True) -> Union[float, np.ndarray]:
    """
    Returns the canonical correlation coefficient of data X and Y.
    If `max` is False, returns all the singular values in decreasing order.
    """

    # Contraction matrix (which singular values are the canonical correlations)   
    M, _, __ = contraction_matrix(X, Y)

    # Singular values decomposition
    _, S, __ = np.linalg.svd(M)
    
    if max:
        return S[0] # S is already sorted in decreasing order
    
    return S

def cca(X: np.ndarray, Y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    TODO
    """

    # Contraction matrix (which singular values are the canonical correlations)   
    M, R_XX_Pud, R_YY_Pud = contraction_matrix(X, Y)

    # Décomposition en valeurs singulières pour avoir les correlations canoniques
    UX, S, UY = np.linalg.svd(M) # Vérifier l'ordre des outputs
    
    # Projection coefficients
    JX = np.linalg.solve(R_XX_Pud, UX[:, 0])
    JY = np.linalg.solve(R_YY_Pud, UY[:, 0])

    # Constraint of sum to 1 over the coefficient
    JX /= abs(JX).sum()
    JY /= abs(JY).sum()

    return JX, JY, S[0]


from .statistics import *
from .resampling import *
from .preprocessing import *
from .ranking import *

__all__ = []
__all__.extend(statistics.__all__)
__all__.extend(resampling.__all__)
__all__.extend(preprocessing.__all__)
__all__.extend(ranking.__all__)



from typing import Optional, Union

import numpy as np
from tqdm import tqdm

from scipy import stats, integrate

__all__ = [
    "prob_higher"
]

def prob_higher(
    mus: np.ndarray,
    sigmas: np.ndarray,
    idx: Optional[int]=None,
    approx: bool=True,
    pbar: bool=False,
) -> Union[np.ndarray, float]:
    """
    Returns the probability
    Ref: https://stats.stackexchange.com/questions/44139/what-is-px-1x-2-x-1x-3-x-1x-n
    """
    if not isinstance(mus, np.ndarray):
        mus = np.array(mus)
    if not isinstance(sigmas, np.ndarray):
        sigmas = np.array(sigmas)

    assert mus.shape == sigmas.shape
    n_vars = mus.size

    def integrand(t: np.ndarray, i: int) -> np.ndarray:
        res = stats.norm.logpdf(t, loc=mus[i], scale=sigmas[i])
        for j in range(n_vars):
            if j != i:
                res += stats.norm.logcdf(t, loc=mus[j], scale=sigmas[j])
        return np.exp(res)
    # We restrict to a bounded interval to prevent large integration errors due to the small domain where the integrand is not zero
    n_sigma = 5

    # In approx mode, only the probabilities of line combinations whose +/- n_sigma_approx intervals overlap are calculated
    if approx:
        i_higher = np.argmax(mus)
        n_sigma_approx = 1.5
        significant = (mus + n_sigma_approx*sigmas) > (mus[i_higher] - n_sigma_approx*sigmas[i_higher])
    else:
        significant = np.ones_like(mus, dtype=bool)

    # If the user wants to compute all probabilities
    if idx is None:
        probs = np.zeros(n_vars)# * np.nan
        _iterable = tqdm(range(n_vars)) if pbar else range(n_vars)
        for i in _iterable:
            if  significant[i]:
                bounds = (mus[i]-n_sigma*sigmas[i], mus[i]+n_sigma*sigmas[i])
                fun = lambda t: integrand(t, i)
                p = integrate.quad(fun, *bounds)[0]
                probs[i] = p
        return probs / np.sum(probs)
    
    # If the user has asked for a particular probability
    bounds = (mus[idx]-n_sigma*sigmas[idx], mus[idx]+n_sigma*sigmas[idx])
    fun = lambda t: integrand(t, idx)
    return integrate.quad(fun, *bounds)[0]



import os
import shutil
import pickle
import itertools as itt
from typing import List, Dict, Any, Tuple, Union, Optional, Sequence, Callable
from copy import deepcopy

import numpy as np
from tqdm import tqdm

from ..stats.statistics import Statistic
from ..stats.resampling import Resampling

from .handler import Handler

__all__ = [
    "ContinuousHandler"
]


class ContinuousHandler(Handler):

    save_path: str

    getter: Callable[[List[str], List[str], Dict[str, Tuple[float, float]]], Tuple[np.ndarray, np.ndarray]]

    stats: Dict[str, Callable]
    resamplings: Dict[str, Resampling]={}


    # Writing access

    def create(
        self,
        x_names: Union[str, Sequence[str]],
        y_names: Union[str, Sequence[str]]
    ):
        """
        Create the statistics directory if not exists as well as the pickle file.
        """
        if not os.path.exists(self.save_path):
            os.makedirs(self.save_path)
        
        path = self.get_filename(x_names, y_names)
        if not os.path.isfile(path):
            with open(path, 'wb') as f:
                pickle.dump([], f)

    def remove(
        self,
        targets: Optional[Sequence[Union[str, Sequence[str]]]],
        variables: Optional[Sequence[Union[str, Sequence[str]]]]
    ) -> None:
        """
        If `targets` and `variables` are both None, remove the whole directory if exists.
        If `targets` and `variables` are both not None, only remove the corresponding pickle file.
        Else raise an error.
        """
        if targets is None and variables is None:
            if os.path.exists(self.save_path):
                shutil.rmtree(self.save_path)
            return

        if (targets is None) ^ (variables is None):
            raise ValueError("targets and variables must be simultaneously None or not None")

        path = self.get_filename(targets, variables)
        if os.path.isfile(path):
            os.remove(path)

    def delete_stats(
        self,
        targs: Union[str, List[str]],
        stats: Union[str, List[str]],
        vars: Optional[Union[str, List[str]]]=None,
    ) -> None:
        if vars is None:
            vars = self.get_available_variables(targs)
        elif isinstance(vars, str):
            vars = [vars]

        for v in vars:
            path = self.get_filename(v, targs)

            print(path)
            with open(path, 'rb') as f:
                d = pickle.load(f)

            if isinstance(stats, str):
                stats = [stats]

            for stat in stats:
                exists = False
                for item in d:
                    exists |= item["stats"].pop(stat, None) is not None 
                
                if exists:
                    print(f"Removing {stat} in {os.path.basename(path)}")

            with open(path, 'wb') as f:
                pickle.dump(d, f)

    def store(
        self,
        x_names: Union[str, List[str]],
        y_names: Union[str, List[str]],
        inputs_dict: Dict[str, Any],
        overwrite: bool=False,
        raise_error: bool=True
    ) -> None:
        """
        Inputs_dict:
        - TODO
        """
        if self.getter is None:
            raise RuntimeError("You must call self.set_getter before calling self.update, self.overwrite or self.store.")

        # Checks that the file is in the expected format.
        inputs_dict = self._check_inputs(
            inputs_dict,
        )

        if isinstance(x_names, str):
            x_names = [x_names]
        if isinstance(y_names, str):
            y_names = [y_names]

        x_names, y_names = sorted(x_names), sorted(y_names)

        # Create directory of not exists
        self.create(x_names, y_names)

        # Load existing data
        path = self.get_filename(x_names, y_names)
        with open(path, 'rb') as f:
            results = pickle.load(f)  # results is a list of dicts

        # Windows loop
        for wins in inputs_dict['windows']:

            wins = wins.copy()
            # for key in ['features', 'bounds', 'scale', 'length', 'points']: TODO
            #     if key not in wins:
            #         wins[key] = [None]
            #     elif not isinstance(wins[key], List):
            #         wins[key] = [wins[key]]

            # We check if the data already exists
            index_windows = self._index_of_data(
                results, value=wins['features'], key="features"
            )

            # We create an entry if it doesn't exists
            if index_windows is None:
                results.append({
                    "features": wins['features'],
                    "stats": {},
                })
                index_windows = -1

            entry = results[index_windows]["stats"]
            for stat in inputs_dict["statistics"]:

                # We pass if the existing statistic exists and that overwrite is False
                if not overwrite and stat in entry:
                    continue
                if stat not in entry:
                    entry.update({stat: {}})

                operator = self.stats[stat] # Callable(ndarray, ndarray) -> float

                self._compute_stat(
                    x_names, y_names, wins,
                    operator, stat,
                    inputs_dict, entry[stat],
                    raise_error=raise_error
                ) # Modify `entry` in-place

            # Save results (update for each ranges iteration)
            with open(self.get_filename(x_names, y_names), 'wb') as f:
                pickle.dump(results, f)

    def _compute_stat(
        self,
        x_names: List[str], y_names: List[str], wins: Dict[str, Any],
        operator: Statistic, stat: str,
        inputs_dict: Dict[str, Any], entry: Dict[str, Any],
        raise_error: bool=True # TODO
    ) -> Dict[str, Any]:
        """
        wins:
        - features: List[str]
        - bounds: List[Tuple[float, float]]
        - bounds_include_windows: bool
        - scale: List[Literal[lin, log]]
        - length: List[float]
        - num_windows: List[float]
        - points: List[int]
        - overlap: List[Union[float, str]]

        Incompatible wins options:
        - length ou num_window
        - points ou overlap
        """

        if 'length' not in wins:
            wins['length'] = [None] * len(wins['features'])
        if 'points' not in wins:
            wins['points'] = [None] * len(wins['features'])

        coords = []
        bounds = []
        for i, ((low, upp), winsize, pts, sc) in enumerate(zip(
            wins['bounds'], wins['length'], wins['points'], wins['scale']
        )):
            if winsize is None:
                winsize = (upp/low)**(1/wins['num_windows'][i]) if sc == 'log'\
                    else 1/wins['num_windows'][i] * (upp-low)

            if pts is None:
                ovp = wins['overlap'][i]
                if isinstance(ovp, str):
                    ovp = ovp.strip()
                    if ovp.endswith("%"):
                        ovp = float(ovp.removesuffix("%"))
                    else:
                        raise ValueError("Incorrect string overlap value (must include the % symbol)")
                    ovp = winsize / ovp

            if sc == 'log':
                padd = np.sqrt(winsize) if wins['bounds_include_windows'] else 1.
                if pts is None:
                    pts = (upp/low * padd**2) / ovp
                    pts = round(pts)
                xticks = np.logspace(
                    np.log10(low*padd), np.log10(upp/padd), pts
                )
                bounds.append((xticks/padd, xticks*padd))
                coords.append(xticks)
            else: # lin or None
                padd = winsize/2 if wins['bounds_include_windows'] else 0.
                if pts is None:
                    pts = (upp-low + 2*padd) / ovp
                    pts = round(pts)
                xticks = np.linspace(
                    low+padd, upp-padd, pts
                )
                bounds.append((xticks-padd, xticks+padd))
                coords.append(xticks)
            
            wins["points"] = [ticks.size for ticks in coords]

        data = np.zeros(wins["points"])
        samples = np.zeros(wins["points"])
        
        if stat in inputs_dict["uncertainty"] and "name" in inputs_dict["uncertainty"][stat]: 
            std = np.zeros(wins['points'])
            name_resampling = inputs_dict["uncertainty"][stat]["name"]
            args = inputs_dict["uncertainty"][stat]["args"]
        else:
            std = None

        pbar = tqdm(
            itt.product(*[range(n) for n in wins['points']]),
            total=np.prod(wins['points']),
            leave=False,
            desc=f"Stat: {stat}, window: " + str(wins['features']).replace("'", "")
        )
        for ii in pbar:
            restrict = {key: [b[0][i], b[1][i]] for i, b, key in zip(ii, bounds, wins['features'])}
            _X, _Y = self.getter(
                x_names, y_names, restrict, inputs_dict['max_samples']
            )

            # Samples
            n = _Y.shape[0]
            samples[ii] = n

            # Computation of value
            if n >= inputs_dict['min_samples']:
                try:
                    data[ii] = operator(_Y, _X)
                except:
                    data[ii] = np.nan
            else:
                data[ii] = np.nan

            if std is None:
                continue

            # Computation of estimator variance
            if n >= inputs_dict['min_samples']:
                try:
                    std[ii] = self.resamplings[name_resampling].compute_sigma(_X, _Y, operator, **args)
                except:
                    std[ii] = np.nan
            else:
                std[ii] = np.nan

        entry.update({
            "coords": tuple(coords),
            "data": data,
            "std": std,
            'samples': samples
        })
        return entry


    # Reading access

    def read(
        self,
        x_names: Union[str, Sequence[str]],
        y_names: Union[str, Sequence[str]],
        wins_features: Union[str, Sequence[str]]
    ) -> Dict[str, Any]:
        # Formatting
        if isinstance(x_names, str):
            x_names = [x_names]
        if isinstance(y_names, str):
            y_names = [y_names]
        if isinstance(wins_features, str):
            wins_features = [wins_features]

        # Load data
        path = self.get_filename(x_names, y_names)
        if not os.path.exists(path):
            raise FileNotFoundError(f"File {path} not exists yet.")
        with open(path, 'rb') as f:
            data = pickle.load(f)  # data is a list of dicts

        # Find the good windows
        names = set(wins_features)
        found = False
        for item in data:
            if set(item["features"]) == names:
                data = item
                found = True
                break
        if not found:
            raise ValueError(f"Windows {names} doesn't exist in data")

        return data

    def get_available_targets(
        self
    ) -> List[List[str]]:
        raise NotImplementedError("TODO")

    def get_available_variables(
        self,
        y_names: Union[None, str, List[str]]
    ):
        raise NotImplementedError("TODO")
        if isinstance(targets, str):
            targets = [targets]
        files = os.listdir(self.save_path)
        files = [f.replace(".pickle", "").split("__", 1) for f in files if f.endswith(".pickle")]
        vars = [s2.split("_") for s1, s2 in files if set(s1.split("_")) == set(targets)]
        return vars

    def get_available_window_features(
        self,
        x_names: Union[str, List[str]],
        y_names: Union[str, List[str]]
    ):
        raise NotImplementedError("TODO")

    def get_available_stats(
        self,
        x_names: Union[str, List[str]],
        y_names: Union[str, List[str]],
        window_features: Union[str, List[str]]
    ):
        raise NotImplementedError("TODO")

    @staticmethod
    def _index_of_data(
        ls: Sequence[Dict[str, Sequence]],
        value: Sequence,
        key: str
    ) -> Optional[int]:
        """
        Order of element and case are important.
        """
        try:
            index = [tuple(item[key]) for item in ls].index(tuple(value))
        except ValueError:
            index = None
        return index
    
    def get_filename(
        self,
        x_names: Union[str, Sequence[str]],
        y_names: Union[str, Sequence[str]]
    ) -> str:
        if isinstance(x_names, str):
            x_names = [x_names]
        if isinstance(y_names, str):
            y_names = [y_names]
        filename = "_".join(sorted(y_names)) + "__" + "_".join(sorted(x_names))
        return os.path.join(self.save_path, filename + '.pickle')

    def _check_inputs(
        self,
        inputs_dict: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        windows:
        - features: str or List[str]
          bounds: List[int, int] or List[List[int, int]]
          bounds_include_windows: bool or List[bool] (optional)
          scale: Literal[linear, log] or List[Literal[linear, log]] (optional)
          length: float or str or List[float or str] (optional, length ^ num_windows)
          num_windows: float or List[float] (optional, length ^ num_windows)
          points: int or List[int] (optional, points ^ overlap)
          overlap: float or str or List[float or str] (optional, points ^ overlap)
            
        min_samples: int (optional)
        max_samples: int (optional)

        statistics: List[str]

        uncertainty (optional):
            str:
                name: str
                args:
                    arg1: Any
                    ...
            ...
        """
        inputs_dict = deepcopy(inputs_dict)

        mandatory_keys = [
            "windows",
            "statistics",
        ]
        for key in mandatory_keys:
            assert key in inputs_dict

        optional_keys = [
            "min_samples",
            "max_samples",
            "uncertainty"
        ]

        for key in inputs_dict:
            if key not in mandatory_keys:
                assert key in optional_keys

        key = "windows"
        if isinstance(inputs_dict[key], Dict):
            inputs_dict[key] = [inputs_dict[key]]
        assert isinstance(inputs_dict[key], List)
        assert all([isinstance(el, Dict) for el in inputs_dict[key]])
        mandatory_window_keys = ["features", "bounds"]
        optional_window_keys = ["bounds_include_windows", "scale"]
        special_window_keys_1 = ["length", "num_windows"]
        special_window_keys_2 = ["points", "overlap"]
        for d in inputs_dict[key]:
            assert set(d.keys()) <= set(mandatory_window_keys\
                + optional_window_keys + special_window_keys_1 + special_window_keys_2)
            for name in mandatory_window_keys:
                assert name in d
            assert sum([name in d for name in special_window_keys_1]) == 1
            assert sum([name in d for name in special_window_keys_2]) == 1

            if isinstance(d["features"], str):
                d["features"] = [d["features"]]
            n_features = len(d["features"])

            assert isinstance(d["features"], List)
            if isinstance(d["bounds"], List) and not isinstance(d["bounds"][0], List):
                d["bounds"] = [d["bounds"]] * n_features
            assert isinstance(d["bounds"], List)
            if "bounds_include_windows" in d:
                if isinstance(d["bounds_include_windows"], bool):
                    d["bounds_include_windows"] = [d["bounds_include_windows"]]
                assert isinstance(d["bounds_include_windows"], List)
            if "scale" in d:
                if isinstance(d["scale"], str):
                    d["scale"] = [d["scale"]] * n_features
                assert isinstance(d["scale"], List)
            if "length" in d:
                if isinstance(d["length"], (float, int)):
                    d["length"] = [d["length"]]
                assert isinstance(d["length"], List)
                assert len(d["length"]) ==  n_features
            if "num_windows" in d:
                if isinstance(d["num_windows"], (float, int)):
                    d["num_windows"] = [d["num_windows"]] * n_features
                assert isinstance(d["num_windows"], List)
            if "points" in d:
                if isinstance(d["points"], int):
                    d["points"] = [d["points"]] * n_features
                assert isinstance(d["points"], List)
            if "overlap" in d:
                if isinstance(d["overlap"], (str, float, int)):
                    d["overlap"] = [d["overlap"]] * n_features
                assert isinstance(d["overlap"], List)

            # Percents
            for name in ["length", "overlap"]:
                if name not in d:
                    continue
                for i, el in enumerate(d[name]):
                    if isinstance(el, str):
                        el = el.strip()
                        assert el[-1] == "%"
                        d[i] = el.strip()
                    assert isinstance(el, (float, int, str))

        key = "min_samples"
        if key not in inputs_dict:
            inputs_dict[key] = 0
        assert isinstance(inputs_dict[key], int)

        key = "max_samples"
        if key not in inputs_dict:
            inputs_dict[key] = None
        assert isinstance(inputs_dict[key], int) or inputs_dict[key] is None

        key = "statistics"
        if isinstance(inputs_dict[key], str):
            inputs_dict[key] = [inputs_dict[key]]
        assert isinstance(inputs_dict[key], List)
        assert all([isinstance(el, str) for el in inputs_dict[key]])

        key = "uncertainty"
        if key not in inputs_dict:
            inputs_dict[key] = {}
        assert isinstance(inputs_dict[key], Dict)
        assert all([isinstance(k, str) and isinstance(d, Dict)\
                    for k, d in inputs_dict[key].items()])
        assert all(["name" in d for d in inputs_dict[key].values()])
        for k in inputs_dict[key]:
            assert inputs_dict[key][k]["name"] in self.resamplings
            if "args" not in inputs_dict[key][k]:
                inputs_dict[key][k] = {}

        return inputs_dict


    # Display

    def __str__(self):
        return f"ContinuousHandler"



from typing import Union

import numpy as np

__all__ = [
    "mse_lower_bound",
    "rmse_lower_bound",
    "corr_to_info",
    "info_to_corr"
]


# Lower bound of MSE
def mse_lower_bound(cond_h: Union[float, np.ndarray], base: float=2) :
    return 1/(2*np.pi*np.e) * np.exp(2*cond_h*np.log(base))

# Lower bound of RMSE
def rmse_lower_bound(cond_h: Union[float, np.ndarray], base: float=2) :
    return np.sqrt(mse_lower_bound(cond_h, base = base))

# Correlation to amount of information
def corr_to_info(rho, base: float=2):
    return - 0.5 * np.log(1 - rho**2) / np.log(base)

# Information to correlation
def info_to_corr(mi, base: float=2):
    return np.sqrt(1 - base**(-2*mi))



from abc import ABC, abstractmethod
from math import floor
from random import shuffle

import numpy as np

from .statistics import Statistic

__all__ = [
    "Resampling",
    "Bootstrapping",
    "Subsampling"
]


class Resampling(ABC):

    def __init__(self):
        pass

    @abstractmethod
    def compute_sigma(
        self,
        variables: np.ndarray,
        targets: np.ndarray,
        stat: Statistic
    ) -> float:
        pass

class Bootstrapping(Resampling):

    def compute_sigma(
        self,
        variables: np.ndarray,
        targets: np.ndarray,
        stat: Statistic,
        n: int=10,
    ) -> float:
        if n < 2:
            return np.nan
        assert variables.shape[0] == targets.shape[0]
        values = np.zeros(n)
        for i in range(n):
            idx = np.random.choice(variables.shape[0], variables.shape[0], replace=True)
            values[i] = stat(variables[idx], targets[idx])
        return np.std(values)

class Subsampling(Resampling):

    def compute_sigma(
        self,
        variables: np.ndarray,
        targets: np.ndarray,
        stat: Statistic,
        n: int=5,
        min_samples: int=20,
        min_subsets: int=5,
        decades: float=2,

    ) -> float:
        if n < 2:
            return np.nan

        assert variables.shape[0] == targets.shape[0]
        N = variables.shape[0]

        max_Ni = floor(N / min_subsets)
        min_Ni = max(min_samples, 10**(-decades) * max_Ni)
        Nis = np.logspace(np.log10(min_Ni), np.log10(max_Ni), n).astype(int)
        Nis = np.unique(Nis)
        if Nis.size < 2:
            return np.nan

        ni_list = []
        sigma2_list = []
        idx = list(range(N))
        for Ni in Nis:
            ni = floor(N / Ni)
            ni_list.append(ni)
            shuffle(idx)
            mis = []
            for i in range(ni):
                sidx = idx[i*Ni:(i+1)*Ni]
                mis.append(stat(
                    variables[sidx],
                    targets[sidx]
                ))
            sigma2_list.append(np.var(mis))

        nis = np.array(ni_list)
        sigma2s = np.array(sigma2_list)

        return np.sqrt( np.sum((nis-1)/nis * sigma2s) / np.sum(nis - 1) )



#!/usr/bin/env python
# Based on the package by Greg Ver Steeg
# See readme.pdf for documentation
# Or go to http://www.isi.edu/~gregv/npeet.html

import numpy as np
from scipy.special import digamma
from sklearn.neighbors import BallTree, KDTree

__all__ = [
    "entropy",
    "centropy",
    "mi"
]


# Nearest neighbors-based estimators

def entropy(
    x: np.ndarray,
    k: int=3,
    base: float=2
) -> float:
    """
    The classic K-L k-nearest neighbor continuous entropy estimator
        x should be a list of vectors, e.g. x = [[1.3], [3.7], [5.1], [2.4]]
        if x is a one-dimensional scalar and we have four samples
    """
    assert x.ndim in (1, 2)
    N = x.shape[0]

    assert k <= N - 1

    if x.ndim == 1:
        x = np.expand_dims(x, 1)
    
    n_elements, n_features = x.shape
    x = _add_noise(x)
    tree = _build_tree(x)
    nn = _query_neighbors(tree, x, k)
    const = digamma(n_elements) - digamma(k) + n_features * np.log(2)

    return (const + n_features * np.log(nn).mean()) / np.log(base)

def centropy(
    x: np.ndarray,
    y: np.ndarray,
    k: int=3, base:
    float=2
) -> float:
    """
    The classic K-L k-nearest neighbor continuous entropy estimator for the
        entropy of X conditioned on Y.
    """
    assert x.ndim in (1, 2)
    assert y.ndim in (1, 2)
    assert x.shape[0] == y.shape[0]
    N = x.shape[0]

    assert k <= N - 1

    if x.ndim == 1:
        x = np.expand_dims(x, 1)
    if y.ndim == 1:
        y = np.expand_dims(y, 1)

    xy = np.column_stack((x, y))
    entropy_union_xy = entropy(xy, k=k, base=base)
    entropy_y = entropy(y, k=k, base=base)

    return entropy_union_xy - entropy_y

def mi(
    x: np.ndarray,
    y: np.ndarray,
    k: int=3,
    base: float=2
) -> float:
    """
    Mutual information of x and y (conditioned on z if z is not None)
    x, y should be a list of vectors, e.g. x = [[1.3], [3.7], [5.1], [2.4]]
    if x is a one-dimensional scalar and we have four samples
    """
    assert x.ndim in (1, 2)
    assert y.ndim in (1, 2)
    assert x.shape[0] == y.shape[0]
    N = x.shape[0]

    assert k <= N - 1

    if x.ndim == 1:
        x = np.expand_dims(x, 1)
    if y.ndim == 1:
        y = np.expand_dims(y, 1)

    # Small additive noise
    x = _add_noise(x)
    y = _add_noise(y)

    # Find nearest neighbors in joint space, p=inf means max-norm
    points = np.column_stack((x, y))
    tree = _build_tree(points)
    dvec = _query_neighbors(tree, points, k)

    # Kraskov formula
    res = digamma(k) - (_avgdigamma(x, dvec) + _avgdigamma(y, dvec)) + digamma(N)
    return res / np.log(base)


## Helpers

def _add_noise(x, intens=1e-10):
    """
    Small noise to break degeneracy.
    """
    return x + intens * np.random.random_sample(x.shape)

def _query_neighbors(tree, x, k):
    return tree.query(x, k=k+1)[0][:, k]

def _count_neighbors(tree, x, r) -> int:
    return tree.query_radius(x, r, count_only=True)

def _avgdigamma(points, dvec) -> float:
    """
    Finds number of neighbors in some radius in the marginal space.
    Returns expectation value of <psi(nx)>.
    """
    tree = _build_tree(points)
    dvec = dvec - 1e-15
    num_points = _count_neighbors(tree, points, dvec)
    return np.mean(digamma(num_points))

def _build_tree(points):
    if points.shape[1] >= 20:
        return BallTree(points, metric='chebyshev')
    return KDTree(points, metric='chebyshev')



import itertools as itt
from abc import ABC, abstractmethod
from typing import List, Dict, Any, Tuple, Union, Optional, Sequence, Callable

import numpy as np

from ..stats import statistics as stt
from ..stats import resampling as rsp

__all__ = [
    "Handler"
]


class Handler(ABC):

    save_path: str

    getter: Callable[[List[str], List[str], Dict[str, Tuple[float, float]]], Tuple[np.ndarray, np.ndarray]]

    stats: Dict[str, Callable]
    resamplings: Dict[str, rsp.Resampling]={}

    def __init__(
        self
    ):
        self.stats = {
            'mi': stt.MI(),
            'condh': stt.Condh(),
            'corr': stt.Corr(),
            'linearinfo': stt.LinearInfo(),
            'linearinforeparam': stt.LinearInfoReparam(),
        }

        self.resamplings = {
            'bootstrapping': rsp.Bootstrapping(),
            'subsampling': rsp.Subsampling(),
        }

        self.save_path = None

        self.getter = None

        self.additional_stats = None
        self.additional_resamplings = None


    # Setter
        
    def set_path(
        self,
        save_path: Optional[str]=None
    ) -> None:
        self.save_path = save_path
        
    def set_getter(
        self,
        getter: Callable[[List[str], List[str], Dict[str, Tuple[float, float]]], Tuple[np.ndarray, np.ndarray]],
    ) -> None:
        assert getter is not None
        self.getter = getter

    def set_additional_stats(
        self,
        additional_stats: Dict[str, stt.Statistic]={},
    ) -> None:
        assert all([isinstance(el, stt.Statistic) for el in additional_stats.values()])
        self.stats.update(additional_stats)

    def set_additional_resamplings(
        self,
        additional_resamplings: Dict[str, Callable]={},
    ) -> None:
        assert all([isinstance(el, rsp.Resampling) for el in additional_resamplings.values()])
        self.stats.update(additional_resamplings)


    # General
        
    @abstractmethod
    def get_filename(
        self,
        targets: Union[str, Sequence[str]],
        variables: Union[str, Sequence[str]]
    ) -> str:
        pass

    @staticmethod
    def _comb(
        ls: Sequence,
        repeats: Union[int, Sequence[int]]
    ) -> List[Sequence]:
        res = []
        for r in repeats:
            res += list(itt.combinations(ls, r))
        return res


    # Writing access

    @abstractmethod
    def create(
        self,
        targets: Union[str, Sequence[str]],
        variables: Union[str, Sequence[str]]
    ):
        pass

    @abstractmethod
    def remove(
        self,
        targets: Optional[Sequence[Union[str, Sequence[str]]]],
        variables: Optional[Sequence[Union[str, Sequence[str]]]]
    ) -> None:
        pass

    @abstractmethod
    def delete_stats(
        self,
        targs: Union[str, List[str]],
        stats: Union[str, List[str]]
    ) -> None:
        pass

    def update(
        self,
        x_names: Union[str, List[str]],
        y_names: Union[str, List[str]],
        inputs_dict: Dict[str, Any],
        **kwargs
    ) -> None:
        """
        Keeps the previous results and only recompute the stats that has never been computed before.
        Note that there is no guarantee that the data used.
        An error is raised if the number of samples is different from that already stored, if any.
        """
        self.store(
            x_names,
            y_names,
            inputs_dict,
            overwrite=True,
            **kwargs
        )

    def overwrite(
        self,
        x_names: Union[str, List[str]],
        y_names: Union[str, List[str]],
        inputs_dict: Dict[str, Any],
        **kwargs
    ) -> None:
        """
        Recomputes the statistics that has already been calculated.
        Stats already calculated but not included in the inputs_dict are kept.
        An error is raised if the number of samples is different from that already stored, if any.
        """
        self.store(
            x_names,
            y_names,
            inputs_dict,
            overwrite=True,
            **kwargs
        )

    @abstractmethod
    def store(
        self,
        inputs_dict: Dict[str, Any],
        overwrite: bool=False
    ):
        pass


    # Reading access

    @abstractmethod
    def read(
        self,
        targs: Union[str, Sequence[str]],
        vars: Union[str, Sequence[str]],
        wins_targs: Union[str, Sequence[str]]
    ) -> Dict[str, Any]:
        pass
    

    @abstractmethod
    def _check_inputs(
        self,
        inputs_dict: Dict[str, Any],
        ref_dict: Dict[str, Any]
    ) -> Dict[str, Any]:
        # TODO
        # raise ValueError("Inputs must match the reference file")
        pass


    # Display

    def overview(self):
        print("Handler")
        print("Save path:", self.save_path)

    @abstractmethod
    def __str__(self):
        pass



from .handler import *
from .discrete_handler import *
from .continuous_handler import *
from .getters import *

__all__ = []
__all__.extend(handler.__all__)
__all__.extend(discrete_handler.__all__)
__all__.extend(continuous_handler.__all__)
__all__.extend(getters.__all__)



import os
import shutil
import json
import itertools as itt
from typing import List, Dict, Any, Tuple, Union, Optional, Sequence, Callable, Iterable
from time import time

import numpy as np
from tqdm import tqdm

from ..stats.statistics import Statistic
from ..stats.resampling import Resampling

from .handler import Handler

__all__ = [
    "DiscreteHandler"
]


class DiscreteHandler(Handler):

    save_path: str

    getter: Callable[[List[str], List[str], Dict[str, Tuple[float, float]]], Tuple[np.ndarray, np.ndarray]]

    stats: Dict[str, Callable]
    resamplings: Dict[str, Resampling]={}

    restrictions: Optional[Dict[str, Dict]]=None


    # Setter

    def set_restrictions(self, d: Dict[str, Dict]) -> None:
        self.restrictions = d


    # General

    def get_filename(
        self,
        y_names: Union[str, Sequence[str]]
    ) -> str:
        if isinstance(y_names, str):
            y_names = [y_names]
        _y_names = "_".join(sorted(y_names))
        return os.path.join(self.save_path, _y_names + '.json')


    # Writing access

    def create(
        self,
        y_names: Union[str, Sequence[str]]
    ):
        """
        Create the statistics directory if not exists as well as the JSON files for features in `y_names`.
        """
        if not os.path.exists(self.save_path):
            os.makedirs(self.save_path)
        
        path = self.get_filename(y_names)
        if not os.path.isfile(path):
            with open(path, 'w', encoding="utf-8") as f:
                json.dump([], f, ensure_ascii=False, indent=4)

    def remove(
        self,
        y_names: Optional[Sequence[Union[str, Sequence[str]]]]
    ) -> None:
        """
        If `y_names` is None, remove the whole directory if exists.
        If `y_names` is not None, only remove the corresponding JSON files.
        """
        if not os.path.isdir(self.save_path):
            raise FileNotFoundError(f"Save directory {self.save_path} does not exist.")


        if y_names is None:
            if any([not file.endswith((".json", ".pickle", ".pkl")) for file in os.listdir(self.save_path)]):
                raise PermissionError(f"Save directory {self.save_path} contains files or directories that are not generated with handlers. You should remove it by hand.")
            shutil.rmtree(self.save_path)
            return

        for tar in y_names:
            path = self.get_filename(tar)
            if os.path.isfile(path):
                os.remove(path)

    def delete_stats(
        self,
        targs: Union[str, List[str]],
        stats: Union[str, List[str]]
    ) -> None:
        path = self.get_filename(targs)
        with open(path, 'r') as f:
            d = json.load(f)  # results is a list of dicts

        if isinstance(stats, str):
            stats = [stats]

        for stat in stats:
            for item1 in d:
                for item2 in item1["stats"]:
                    item3 = item2["stats"]
                    for key in [f'{stat}', f'{stat}-time', f'{stat}-boot']:
                        item3.pop(key, None)

        with open(path + ".tmp", 'w', encoding="utf-8") as f:
            json.dump(d, f, ensure_ascii=False, indent=4)
        os.rename(path + ".tmp", path)

    def store(
        self,
        x_names: Union[str, List[str], Iterable[List[str]]],
        y_names: Union[str, List[str]],
        inputs_dict: Dict[str, Any],
        overwrite: bool=False,
        iterable_x: bool=False,
        save_every: int=1,
        progress_bar: bool=True,
        total_iter: int=None,
        raise_error: bool=True
    ) -> None:
        """
        Inputs_dict:
        - TODO
        """

        # Checks that the file is in the expected format.
        self._check_inputs(
            inputs_dict,
        )

        if isinstance(y_names, str):
            y_names = [y_names]

        y_names = sorted(y_names)

        # Create directory of not exists
        self.create(y_names)

        # Load existing data
        path = self.get_filename(y_names)
        with open(path, 'r') as f:
            results = json.load(f)  # results is a list of dicts

        # Variables loop
        if iterable_x:
            assert isinstance(x_names, Iterable) and not isinstance(x_names, str)
        else:
            x_names = [x_names]

        pbar = tqdm(
            x_names,
            desc=str(y_names).replace("'", ""),
            total=total_iter,
            disable=not progress_bar
        )
        for it, _x_names in enumerate(pbar, 1):
            if isinstance(_x_names, str):
                _x_names = [_x_names]    

            _lines = list(set(_x_names))
            pbar.set_postfix({'x': str(_lines).replace("'", "")})

            _x_names = list(set(_x_names))            
            _x_names = sorted(_x_names)

            # We check if the combination of variables already exists
            index_x = self._index_of(
                results, value=_x_names, key="x_names"
            )

            # We create an entry if it doesn't exists
            if index_x is None:
                results.append({
                    "x_names": _x_names,
                    "stats": [],
                })
                index_x = -1

            # Ranges loop
            for restr in inputs_dict["restrictions"]: # TODO: "restrictions" field can be None

                # We check if ranges already exists
                index_ranges = self._index_of_ranges(
                    results[index_x]["stats"], value=restr, key="restriction"
                )

                # We create an entry if it doesn't exists
                if index_ranges is None:
                    results[index_x]["stats"].append({
                        "restriction": restr,
                        "stats": {},
                    })
                    index_ranges = -1

                # Ranges restriction
                if self.restrictions is not None and restr is not None:
                    restrict_dict = self.restrictions[restr]
                elif self.restrictions is not None:
                    raise ValueError(f"self.restriction must not be None when the restriction asked is not None (here {restr}). Consider using set_restrictions to load the dictionnary.")
                else:
                    restrict_dict = {}

                _X, _Y = self.getter(
                    _x_names, y_names, restrict_dict, inputs_dict.get('max_samples')
                )

                entry = results[index_x]["stats"][index_ranges]["stats"]
                for stat in inputs_dict["statistics"]:
                    if overwrite and stat in entry:
                        entry.pop(stat, None)
                if set(entry.keys()) <= {"samples"}: # If no keys or only "samples"
                    entry.pop("samples", None)

                for stat in inputs_dict["statistics"]:

                    if stat in entry:
                        continue
                    entry.update({stat: {}})

                    operator = self.stats[stat]

                    self._compute_stat(
                        _X, _Y,
                        operator, stat,
                        inputs_dict, entry[stat],
                        raise_error=raise_error
                    ) # Modify `entry` in-place
                
                samples = _Y.shape[0]
                prev_samples = entry.get("samples")
                if prev_samples is not None and samples != samples:
                    raise ValueError("Old and new number of samples are differents.")

                entry.update({
                    "samples": samples
                })

            # Save results
            
            if it % save_every == 0:
                path = self.get_filename(y_names)
                with open(path + ".tmp", 'w', encoding="utf-8") as f:
                    json.dump(results, f, ensure_ascii=False, indent=4)
                os.rename(path + ".tmp", path)

        # Final save
        
        path = self.get_filename(y_names)
        with open(path + ".tmp", 'w', encoding="utf-8") as f:
            json.dump(results, f, ensure_ascii=False, indent=4)
        os.rename(path + ".tmp", path)

    def _compute_stat(
        self,
        X: np.ndarray, Y: np.ndarray,
        operator: Statistic, stat: str,
        inputs_dict: Dict[str, Any], entry: Dict[str, Any],
        raise_error: bool=True
    ) -> Dict[str, Any]:
        """
        Modifies in-place the `entry` dictionnary.
        Adds or changes the three following entries of `entry`:
        - {stat}: computed only if current value is None.
            If error, the default value is None.
        - {stat}-time: computed only if new value of {stat} is not None.
        - {stat}-boot: computed only if new value of {stat} is not None
            and if `inputs_dict["bootstrapping"][stat]` is not None.
            If current value is None, then the entry is replaced by the new array.
            If current value is not None, then we compute as many values as necessary to reach
            `inputs_dict["bootstrapping"][stat]` total values.

        Note: if the available number of samples is lower than `inputs_dict["min_samples"]`,
        then all three entries are set to None.
        """
        # Samples
        samples = Y.shape[0]
        if samples <= inputs_dict["min_samples"]:
            return {
                "value": None,
                "time": None,
                "std": None
            }

        # Simple computation
        try:
            start = time()
            value = operator(X, Y)
            end = time()
            entry.update({
                "value": value,
                "time": end-start,
            })
        except Exception as e:
            if raise_error:
                raise e
            entry.update({
                "value": None,
                "time": None,
                "std": None
            })
            return entry

        # Uncertainty
        try:
            d = inputs_dict["uncertainty"][stat]
            name, args = d["name"], d["args"]
        except Exception as e:
            if raise_error:
                raise e
            entry["std"] = None
            return

        try:
            std = self.resamplings[name].compute_sigma(X, Y, operator, **args)
        except Exception as e:
            if raise_error:
                raise e
            std = None
        entry.update({
            "std": std
        })

        return entry

    # Reading access

    def _get_variables_content(
        self,
        x_names: List[str],
        data: List[Dict[str, Any]]
    ) -> Optional[List[Dict]]:
        _x_names = set(x_names)
        for _item in data:
            if set(_item["x_names"]) == _x_names:
                item = _item["stats"]
                return item
        return None
    
    def _get_restriction_content(
        self,
        restr: str,
        data: List[Dict[str, Any]]
    ) -> Dict[str, Any]:
        for _item in data:
            if _item["restriction"] == restr:
                item = _item["stats"]
                return item
        return None

    def read(
        self,
        x_names: Union[str, List[str], Iterable[List[str]]],
        y_names: Union[str, List[str]],
        restr: str,
        iterable_x: bool=False,
        default: Any="raise"
    ):
        if isinstance(x_names, str):
            x_names = [x_names]

        if isinstance(y_names, str):
            y_names = [y_names]
        assert isinstance(y_names, List)

        # Load data
        path = self.get_filename(y_names)
        if not os.path.exists(path):
            raise FileNotFoundError(f"File {path} not exists yet.")
        with open(path, 'r') as f:
            data = json.load(f)  # results is a list of dicts

        if not iterable_x:
            x_names = [x_names]

        item_list = []
        for _x_names in x_names:
            if isinstance(_x_names, str):
                _x_names = [_x_names]

            # Find the good set of variables
            item = self._get_variables_content(_x_names, data)
            if item is None:
                if default == "raise":
                    msg = f"Variables {_x_names} doesn't exist in data."
                    if any([isinstance(el, (List, Tuple)) for el in _x_names]):
                        msg += f" It seems that you provide an Iterable, did you missed setting the `iterable_x` flag?"
                    raise ValueError(msg)
                item_list.append(default)
                continue

            # Find the good restriction
            item = self._get_restriction_content(restr, item)
            if item is None:
                if default == "raise":
                    raise ValueError(f"Restriction of data {restr} doesn't exist in data")
                item = default

            # Store value
            item_list.append(item)

        if iterable_x:
            return item_list
        return item_list[0]
    
    def get_available_targets(
        self
    ) -> List[List[str]]:
        """
        TODO
        """
        filenames = [
            f.replace(".json", "") for f in os.listdir(self.save_path) if f.endswith(".json")
        ]
        return [
            f.split("_") for f in filenames
        ]

    def get_available_variables(
        self,
        targets: Union[str, List[str]],
    ) -> List[List[str]]:
        """
        TODO
        """
        # Load data
        path = self.get_filename(targets)
        if not os.path.exists(path):
            raise FileNotFoundError(f"File {path} not exists yet.")
        with open(path, 'r') as f:
            data = json.load(f)  # results is a list of dicts

        return [item["x_names"] for item in data]
    
    def get_available_restrictions(
        self,
        targets: Union[str, List[str]],
        variables: Union[str, List[str]]
    ) -> List[str]:
        """
        TODO
        """
        if isinstance(targets, str):
            targets = [targets]
        if isinstance(variables, str):
            variables = [variables]

        # Load data
        path = self.get_filename(targets)
        if not os.path.exists(path):
            raise FileNotFoundError(f"File {path} not exists yet.")
        with open(path, 'r') as f:
            data = json.load(f)  # results is a list of dicts

        # Get variables content
        content = self._get_variables_content(variables, data)

        return [item["restriction"] for item in content]

    def get_available_stats(
        self,
        targets: Union[str, List[str]],
        variables: Union[str, List[str]],
        restriction: str
    ) -> List[str]:
        if isinstance(targets, str):
            targets = [targets]
        if isinstance(variables, str):
            variables = [variables]

        # Load data
        path = self.get_filename(targets)
        if not os.path.exists(path):
            raise FileNotFoundError(f"File {path} not exists yet.")
        with open(path, 'r') as f:
            data = json.load(f)  # results is a list of dicts

        # Get variables content
        content = self._get_variables_content(variables, data)
        if content is None:
            raise ValueError(f"Invalid variables {variables} for targets {targets}")

        for entry in content:
            if entry["restriction"] == restriction:
                return list(entry["stats"].keys())
        
        raise ValueError(f"Invalid restriction {restriction} for targets {targets} and variables {variables}")

    @staticmethod
    def _index_of(
        ls: Sequence[Dict[str, Sequence]],
        value: Sequence,
        key: str
    ) -> Optional[int]:
        """
        Order of element and case are important.
        """
        try:
            index = [tuple(item[key]) for item in ls].index(tuple(value))
        except ValueError:
            index = None
        return index

    @staticmethod
    def _index_of_ranges(
        ls: Sequence[Dict[str, Dict]],
        value: Sequence,
        key: str
    ) -> Optional[int]:
        """
        Order of element and case are important.
        """
        try:
            index = [item[key] for item in ls].index(value)
        except ValueError:
            index = None
        return index

    def _check_inputs(
        self,
        inputs_dict: Dict[str, Any],
    ) -> None:

        # TODO
        pass
                

    # Display

    def __str__(self):
        return f"DiscreteHandler"



from abc import ABC, abstractmethod

import numpy as np
from scipy.special import erfinv

from .entropy_estimators import mi, centropy
from .canonical_estimators import canonical_corr
from .info_theory import corr_to_info

__all__ = [
    "Statistic",
    "MI",
    "Condh",
    "Corr",
    "LinearInfo",
    "LinearInfoReparam"
]

class Statistic(ABC):
    """
    Abstract class for all statistics used in procedure
    Can be used to defined your own.
    """

    def __init__(self):
        pass

    @abstractmethod
    def __call__(
        self,
        variables: np.ndarray,
        targets: np.ndarray
    ):
        pass

class MI(Statistic):
    """
    TODO
    """

    def __call__(
        self,
        variables: np.ndarray,
        targets: np.ndarray
    ):
        _variables = self.marginaly_gaussianize(variables)
        _targets = self.marginaly_gaussianize(targets)

        return mi(_variables, _targets)
    
    @staticmethod
    def gaussianize(x):
        assert x.ndim == 1
        n = x.size
        order = np.argsort(np.argsort(x))
        r = np.arange(n) + 1
        return np.sqrt(2) * erfinv((2*r[order] - (n+1))/n)

    @staticmethod
    def marginaly_gaussianize(x):
        if x.ndim == 1:
            return MI.gaussianize(x)
        return np.column_stack([
            MI.gaussianize(x[:, i]) for i in range(x.shape[1])
        ])

class Condh(Statistic):
    """
    TODO
    """

    def __call__(
        self,
        variables: np.ndarray,
        targets: np.ndarray
    ):
        _variables = variables + np.random.normal(0, 1e-3, variables.shape)
        _targets = targets + np.random.normal(0, 1e-3, targets.shape)
        return centropy(_targets, _variables)

class Corr(Statistic):
    """
    TODO
    """

    def __call__(
        self,
        variables: np.ndarray,
        targets: np.ndarray
    ):
        return canonical_corr(variables, targets, max=True).item()

class LinearInfo(Statistic):
    """
    TODO
    """

    def __call__(
        self,
        variables: np.ndarray,
        targets: np.ndarray
    ):
        rhos = canonical_corr(variables, targets, max=False)
        return np.sum(corr_to_info(rhos))

class LinearInfoReparam(Statistic):
    """
    TODO
    """

    def __call__(
        self,
        variables: np.ndarray,
        targets: np.ndarray
    ):
        _variables = MI.marginaly_gaussianize(variables)
        _targets = MI.marginaly_gaussianize(targets)

        rhos = canonical_corr(_variables, _targets, max=False)
        return np.sum(corr_to_info(rhos))



import numpy as np

__all__ = [
    "break_degeneracy"
]


def break_degeneracy(
    data: np.ndarray
) -> np.ndarray:
    """
    Measures the sample step and add an adequate noise to break degeneracy.
    data: np.ndarray of shape (N, dim)
    """
    diffs = np.diff(np.sort(data, axis=0), axis=0)
    diffs = np.where(diffs==0, np.nan, diffs)
    
    steps = np.nanmin(diffs, axis=0)
    
    datac = (np.random.random(size=data.shape)-0.5) * steps
    return datac
